@@ -1,7 +1,7 @@
 /* lr - a better recursive ls/find */
 
 /*
- * Copyright (C) 2015-2017 Leah Neukirchen <purl.org/net/chneukirchen>
+ * Copyright (C) 2015-2018 Leah Neukirchen <purl.org/net/chneukirchen>
  * Parts of code derived from musl libc, which is
  * Copyright (C) 2005-2014 Rich Felker, et al.
  *
@@ -88,6 +88,7 @@ static int Dflag;
 static int Hflag;
 static int Lflag;
 static int Qflag;
+static int Pflag;
 static int Uflag;
 static int Xflag;
 static int hflag;
@@ -1055,7 +1056,7 @@ username(uid_t uid)
 
 	if (name)
 		return name;
-	
+
 	struct passwd *p = getpwuid(uid);
 	if (p) {
 		if ((int)strlen(p->pw_name) > uwid)
@@ -1064,7 +1065,7 @@ username(uid_t uid)
 		users = idtree_insert(users, uid, name);
 		return name;
 	}
-	
+
 	return strid(uid);
 }
 
@@ -1630,6 +1631,23 @@ color_size_on(off_t s)
 	fg256(c);
 }
 
+static void
+print_comma(int len, intmax_t i)
+{
+	char buf[64];
+	char *s = buf + sizeof buf;
+
+	*--s = 0;
+	while (1) {
+		*--s = (i % 10) + '0'; if ((i /= 10) <= 0) break;
+		*--s = (i % 10) + '0'; if ((i /= 10) <= 0) break;
+		*--s = (i % 10) + '0'; if ((i /= 10) <= 0) break;
+		*--s = ',';
+	}
+
+	printf("%*.*s", len+len/3, len+len/3, s);
+}
+
 static void
 print_human(intmax_t i)
 {
@@ -1653,25 +1671,98 @@ print_human(intmax_t i)
 
 }
 
+// Decode one UTF-8 codepoint into cp, return number of bytes to next one.
+// On invalid UTF-8, return -1, and do not change cp.
+// Invalid codepoints are not checked.
+//
+// This code is meant to be inlined, if cp is unused it can be optimized away.
+static int
+u8decode(const char *cs, uint32_t *cp)
+{
+	const uint8_t *s = (uint8_t *)cs;
+
+	if (*s == 0)   { *cp = 0; return 0; }
+	if (*s < 0x80) { *cp = *s; return 1; }
+	if (*s < 0xc2) { return -1; }  //cont+overlong
+	if (*s < 0xe0) { *cp = *s & 0x1f; goto u2; }
+	if (*s < 0xf0) {
+		if (*s == 0xe0 && (s[1] & 0xe0) == 0x80) return -1; //overlong
+		if (*s == 0xed && (s[1] & 0xe0) == 0xa0) return -1; //surrogate
+		*cp = *s & 0x0f; goto u3;
+	}
+	if (*s < 0xf5) {
+		if (*s == 0xf0 && (s[1] & 0xf0) == 0x80) return -1; //overlong
+		if (*s == 0xf4 && (s[1] > 0x8f)) return -1; //too high
+		*cp = *s & 0x07; goto u4;
+	}
+	return -1;
+
+u4:	if ((*++s & 0xc0) != 0x80) return -1;  *cp = (*cp << 6) | (*s & 0x3f);
+u3:	if ((*++s & 0xc0) != 0x80) return -1;  *cp = (*cp << 6) | (*s & 0x3f);
+u2:	if ((*++s & 0xc0) != 0x80) return -1;  *cp = (*cp << 6) | (*s & 0x3f);
+	return s - (uint8_t *)cs + 1;
+}
+
 static void
 print_shquoted(const char *s)
 {
-	if (!Qflag || !strpbrk(s, "\001\002\003\004\005\006\007\010"
-	                          "\011\012\013\014\015\016\017\020"
-	                          "\021\022\023\024\025\026\027\030"
-	                          "\031\032\033\034\035\036\037\040"
-	                          "`^#*[]=|\\?${}()'\"<>&;\177")) {
+	uint32_t ignored;
+	int l;
+
+	const char *t;
+	int esc = 0;
+	for (t = s; *t; ) {
+		if ((unsigned char)*t <= 32 ||
+		    strchr("`^#*[]=|\\?${}()'\"<>&;\177", *t)) {
+			esc = 1;
+			break;
+		} else {
+			if ((l = u8decode(t, &ignored)) < 0) {
+				esc = 1;
+				break;
+			}
+			t += l;
+		}
+	}
+
+	if (!Qflag || !esc) {
 		printf("%s", s);
 		return;
 	}
 
-	putchar('\'');
-	for (; *s; s++)
-		if (*s == '\'')
-			printf("'\\''");
-		else
-			putchar(*s);
-	putchar('\'');
+	if (Pflag) {
+		printf("$'");
+		for (; *s; s++)
+			switch (*s) {
+			case '\a': printf("\\a"); break;
+			case '\b': printf("\\b"); break;
+			case '\e': printf("\\e"); break;
+			case '\f': printf("\\f"); break;
+			case '\n': printf("\\n"); break;
+			case '\r': printf("\\r"); break;
+			case '\t': printf("\\t"); break;
+			case '\v': printf("\\v"); break;
+			case '\\': printf("\\\\"); break;
+			case '\'': printf("\\\'"); break;
+			default:
+				if ((unsigned char)*s < 32 ||
+				    (l = u8decode(s, &ignored)) < 0) {
+					printf("\\%03o", (unsigned char)*s);
+				} else {
+					printf("%.*s", l, s);
+					s += l-1;
+				}
+			}
+		putchar('\'');
+	} else {
+		putchar('\'');
+		for (; *s; s++)
+			if (*s == '\'')
+				printf("'\\''");
+			else
+				putchar(*s);
+		putchar('\'');
+	}
 }
 
 void
@@ -1831,6 +1922,7 @@ hyperlink_off()
 void
 print_format(struct fileinfo *fi)
 {
+	int c, v;
 	char *s;
 	for (s = format; *s; s++) {
 		if (*s == '\\') {
@@ -1842,8 +1934,32 @@ print_format(struct fileinfo *fi)
 			case 'r': putchar('\r'); break;
 			case 't': putchar('\t'); break;
 			case 'v': putchar('\v'); break;
-			case '0': putchar('\0'); break;
-			// TODO: \NNN
+			case '0': case '1': case '2': case '3':
+			case '4': case '5': case '6': case '7':
+				for (c = 3, v = 0;
+				     c > 0 && *s >= '0' && *s <= '7';
+				     s++, c--) {
+					v = (v << 3) + ((*s) - '0');
+				}
+				s--;
+				putchar(v & 0xff);
+				break;
+			case 'x':
+				s++;
+				for (c = 2, v = 0;
+				     c > 0 && isxdigit((unsigned char)*s);
+				     s++, c--) {
+					v <<= 4;
+					if (*s >= 'A' && *s <= 'F')
+						v += *s - 'A' + 10;
+					else if (*s >= 'a' && *s <= 'f')
+						v += *s - 'a' + 10;
+					else
+						v += *s - '0';
+				}
+				s--;
+				putchar(v);
+				break;
 			default: putchar(*s);
 			}
 			continue;
@@ -1857,8 +1973,12 @@ print_format(struct fileinfo *fi)
 		case 's':
 			if (!hflag) {
 				color_size_on(fi->sb.st_size);
-				printf("%*jd", intlen(maxsize),
-				    (intmax_t)fi->sb.st_size);
+				if (Gflag)
+					print_comma(intlen(maxsize),
+					    (intmax_t)fi->sb.st_size);
+				else
+					printf("%*jd", intlen(maxsize),
+					    (intmax_t)fi->sb.st_size);
 				fgdefault();
 				break;
 			}
@@ -2294,9 +2414,9 @@ main(int argc, char *argv[])
 
 	setlocale(LC_ALL, "");
 
-	while ((c = getopt(argc, argv, "01ABC:DFGHLQST:UXde:f:lho:st:x")) != -1)
+	while ((c = getopt(argc, argv, "01ABC:DFGHLQPST:UXde:f:lho:st:x")) != -1)
 		switch (c) {
-		case '0': format = zero_format; input_delim = 0; Qflag = 0; break;
+		case '0': format = zero_format; input_delim = 0; Qflag = Pflag = 0; break;
 		case '1': expr = chain(parse_expr("depth > 0 ? prune : print"), EXPR_AND, expr); break;
 		case 'A': expr = chain(expr, EXPR_AND, parse_expr("name =~ \"^\\.\" && path != \".\" ? prune : print")); break;
 		case 'B': Bflag++; Dflag = 0; Uflag = 0; need_stat++; break;
@@ -2316,6 +2436,7 @@ main(int argc, char *argv[])
 		case 'H': Hflag++; break;
 		case 'L': Lflag++; break;
 		case 'Q': Qflag++; break;
+		case 'P': Pflag++; Qflag++; break;
 		case 'S': Qflag++; format = stat_format; break;
 		case 'T': Tflag = timeflag(optarg); break;
 		case 'U': Uflag++; Bflag = 0; break;
@@ -2334,7 +2455,7 @@ main(int argc, char *argv[])
 		case 'x': xflag++; break;
 		default:
 			fprintf(stderr,
-"Usage: %s [-0|-F|-l [-TA|-TC|-TM]|-S|-f FMT] [-B|-D] [-H|-L] [-1AGQdhsx]\n"
+"Usage: %s [-0|-F|-l [-TA|-TC|-TM]|-S|-f FMT] [-B|-D] [-H|-L] [-1AGPQdhsx]\n"
 "          [-U|-o ORD] [-e REGEX]* [-t TEST]* [-C [COLOR:]PATH]* PATH...\n", argv0);
 			exit(2);
 		}
@@ -2347,6 +2468,8 @@ main(int argc, char *argv[])
 		if (Xflag == 1)
 			Xflag = 0;
 	}
+	if (getenv("NO_COLOR"))
+		Gflag = 0;
 
 	analyze_format();
 	if (Uflag) {
